#Function for estimate ET0###############################

fun.ETPHargreaves=function(tmax,tmin,latitude,CCoef=0.0135,CT=17.78,Rs=NULL,day,KT=0.17,HE=0.5,Corr=F,a=0,b=1){
  
  #libraries
  
  if (!require(lubridate)) install.packages('lubridate',dep=T)
  library(lubridate)
  
  tavg = (tmax + tmin) / 2
  mjamm = 0.408
  
  if(is.null(Rs) == TRUE) {
    Gsc = 0.0820
    djulian = yday(day)
    dyear = yday(as.Date(paste0(substr(day,1,4),"-12-31")))
    dr = 1 + 0.033 * cos(2 * pi * djulian / dyear)
    omega = latitude * pi / 180
    delta = 0.409 * sin(((2 * pi) / dyear) * djulian - 1.39)
    ws = acos(-tan(omega) * tan(delta))
    Ra = ((24*60) / pi) * Gsc * dr * (ws * sin(omega) * sin(delta) + cos(omega) * cos(delta) * sin(ws))
    Ra = Ra * mjamm
    Rs = Ra * KT * (tmax - tmin)^HE
  }	else {
    Rs = Rs * mjamm
  }
  
  ET0h = CCoef * Rs * (tavg + CT)
  
  if(Corr == T){
    ET0h = a + b * ET0h
  }
  
  ET0h = data.frame(ET0h,Rs)
  return(ET0h)
  
}

#Function for multi-model ensemble methods###############################

fun.ensembles=function(datahist,datarcp1,datarcp2=NULL,ref,date,CalIni,CalEnd,TestIni,TestEnd,modhist,modensembles,groupK,parallel=T,numcores=1,est.caltest = F){
  
  #libraries
  
  if (!require(GeomComb)) install.packages('GeomComb',dep=T)
  library(GeomComb)
  
  if (!require(BMS)) install.packages('BMS',dep=T)
  library(BMS)
  
  if (!require(caret)) install.packages('caret',dep=T)
  library(caret)
  
  if (!require(randomForest)) install.packages('randomForest',dep=T)
  library(randomForest)
  
  if (!require(kernlab)) install.packages('kernlab',dep=T)
  library(kernlab)

  if(parallel==T){
    if (!require(doParallel)) install.packages('doParallel',dep=T)
    library(doParallel)
  }
  
  if (!require(hydroGOF)) install.packages('hydroGOF',dep=T)
  library(hydroGOF)  

  #Names of ensembles
  namesensembles = c("SA","MED","TA","OLS","LAD","CLS","EIG1","EIG2","BMA","RF","SVR")

  #Data frames to evaluate the models
  validation = validation_cal = validation_test = data.frame(Model=c(modhist,namesensembles),RMSE=NA,PBIAS=NA,NSE=NA,R2=NA,KGE=NA)
  
  models = models_cal = list()

  #order data:
  datahist=datahist[order(datahist[,date]),]
  datarcp1=datarcp1[order(datarcp1[,date]),]
  
  if(is.null(datarcp2) == F){
    datarcp2=datarcp2[order(datarcp2[,date]),]
  }

  #Divide data cal-test
  matrix_cal = datahist[datahist[,date] >= CalIni & datahist[,date] <= CalEnd,]
  matrix_test = datahist[datahist[,date] >= TestIni & datahist[,date] <= TestEnd,]

  #1. Evaluate individual models--------------------------------------------
  
  for(model in modhist){
    #cal
    validation_cal[validation_cal$Model==model,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[datahist[,date] >= CalIni & datahist[,date] <= CalEnd,ref],sim=datahist[datahist[,date] >= CalIni & datahist[,date] <= CalEnd,model],digits=5)[c(4,6,9,17,19)]
    #test
    validation_test[validation_test$Model==model,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[datahist[,date] >= TestIni & datahist[,date] <= TestEnd,ref],sim=datahist[datahist[,date] >= TestIni & datahist[,date] <= TestEnd,model],digits=5)[c(4,6,9,17,19)]
    #complete
    validation[validation$Model==model,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[,ref],sim=datahist[,model],digits=5)[c(4,6,9,17,19)]
  }
  
  #2. Simple, regression and geometric ensembles------------------------------
  
  #combination object for calibration-test
  combination_cal = foreccomb(observed_vector=matrix_cal[,ref], prediction_matrix=matrix_cal[,modensembles], newobs = matrix_test[,ref],newpreds = matrix_test[,modensembles], byrow = FALSE, na.impute = T, criterion = "RMSE")

  #and for complete data
  combination = foreccomb(observed_vector=datahist[,ref], prediction_matrix=as.matrix(datahist[,modensembles]), byrow = FALSE, na.impute = T, criterion = "RMSE")

  #2.1. Simple ensembles----------------------
  #SA
  models_cal$SA = comb_SA(combination_cal)
  models$SA = comb_SA(combination)
  #MED
  models_cal$MED = comb_MED(combination_cal)
  models$MED = comb_MED(combination)
  #TA
  models_cal$TA = comb_TA(combination_cal)
  models$TA = comb_TA(combination)
  
  #2.2. Geometric based ensembles-------------
  #EIG1
  models_cal$EIG1 = comb_EIG1(combination_cal)
  models$EIG1 = comb_EIG1(combination)
  #EIG2
  models_cal$EIG2 = comb_EIG2(combination_cal)
  models$EIG2 = comb_EIG2(combination) 

  #2.3. Regression based ensembles------------
  #OLS
  models_cal$OLS = comb_OLS(combination_cal)
  models$OLS = comb_OLS(combination)
  #LAD
  models_cal$LAD = comb_LAD(combination_cal)
  models$LAD = comb_LAD(combination)
  #CLS
  models_cal$CLS = comb_CLS(combination_cal)
  models$CLS = comb_CLS(combination)
  
  #2.4. BMA-----------------------------------
  models_cal$BMA = bms(matrix_cal[,c(ref,modensembles)], burn=2000,iter = 100000, mcmc="bd", g="UIP", mprior="uniform", nmodel=2000,user.int=FALSE)
  models$BMA = bms(datahist[,c(ref,modensembles)], burn=2000,iter = 100000, mcmc="bd", g="UIP", mprior="uniform", nmodel=2000,user.int=FALSE)
  
  #3. MACHINE LEARNING ENSEMBLES------------------------------------------------

  #Define index for calibration
  g_unique_cal <- unique(matrix_cal[,groupK])
  g_folds_cal <- 1:length(g_unique_cal)
  index2_cal <- split(seq_along(matrix_cal[,groupK]), g_folds_cal[match(matrix_cal[,groupK], g_unique_cal)])
  names(index2_cal) <- paste0("Fold", gsub(" ", "0",format(seq_along(index2_cal))))
  index2_cal=lapply(index2_cal, function(z) seq_along(matrix_cal[,groupK])[-z])
  
  #Define index for complete data
  g_unique <- unique(datahist[,groupK])
  g_folds <- 1:length(g_unique)
  index2 <- split(seq_along(datahist[,groupK]), g_folds[match(datahist[,groupK], g_unique)])
  names(index2) <- paste0("Fold", gsub(" ", "0",format(seq_along(index2))))
  index2=lapply(index2, function(z) seq_along(datahist[,groupK])[-z])
  
  #Enable parallel option
  if(parallel == T){
    cl1 = makeCluster(numcores)
    registerDoParallel(cl1)
    #
    ncvCtrl_cal = trainControl(method = "cv", number=length(unique(matrix_cal[,groupK])),index=index2_cal,search='grid',allowParallel = TRUE,selectionFunction = "tolerance")
    ncvCtrl = trainControl(method = "cv", number=length(unique(datahist[,groupK])),index=index2,search='grid',allowParallel = TRUE,selectionFunction = "tolerance")
  }else{
    ncvCtrl_cal = trainControl(method = "cv", number=length(unique(matrix_cal[,groupK])),index=index2_cal,search='grid',allowParallel = FALSE,selectionFunction = "tolerance")
    ncvCtrl = trainControl(method = "cv", number=length(unique(datahist[,groupK])),index=index2,search='grid',allowParallel = FALSE,selectionFunction = "tolerance")
  }

  #3.1. RF-----------------------------------
  
  tunegrid = expand.grid(.mtry = 1:length(modensembles))
  
  #Cal-val model
  
  mod.rftune_cal = train(x = matrix_cal[,modensembles],y=as.vector(matrix_cal[,ref]),method = "rf",ntree=1000,importance=T,tuneGrid=tunegrid,trControl = ncvCtrl_cal)
  #estimate final model based on calibrated parameter
  models_cal$RF = randomForest(x = matrix_cal[,modensembles],y=as.vector(matrix_cal[,ref]),mtry=mod.rftune_cal$bestTune$mtry,ntree=1000,importance=T,corr.bias=T)
  
  #Complete model
  
  mod.rftune = train(x = datahist[,modensembles],y=as.vector(datahist[,ref]),method = "rf",ntree=1000,importance=T,tuneGrid=tunegrid,trControl = ncvCtrl)
  #estimate final model based on calibrated parameter
  models$RF = randomForest(x = datahist[,modensembles],y=as.vector(datahist[,ref]),mtry=mod.rftune$bestTune$mtry,ntree=1000,importance=T,corr.bias=T)  
  
  #Evaluate relative variable importance
  imp.rf = caret::varImp(mod.rftune)$importance
  
  #3.2. SVR-----------------------------------
  
  #Cal-val model

  #estimate sigma and define grid:
  sigmas_cal = kernlab::sigest(as.matrix(matrix_cal[,modensembles]), na.action = na.omit, scaled = TRUE)
  sigma_cal = mean(as.vector(sigmas_cal[-2]))
  tunegrid_cal = expand.grid(sigma=sigma_cal,C = c(0.75, 1, 2, 4,8,16,32))
  
  mod.svmtune_cal = train(x = matrix_cal[,modensembles],y=as.vector(matrix_cal[,ref]),method = "svmRadial", preProcess = c("center","scale"),importance=T,tuneGrid=tunegrid_cal,trControl = ncvCtrl_cal)
  #estimate final model based on calibrated parameter
  models_cal$SVR = ksvm(x = as.matrix(matrix_cal[,modensembles]),y=as.vector(matrix_cal[,ref]),scaled=T,cross=5,C=mod.svmtune_cal$bestTune$C)

  #Complete model
  
  #estimate sigma and define grid:
  sigmas = kernlab::sigest(as.matrix(datahist[,modensembles]), na.action = na.omit, scaled = TRUE)
  sigma = mean(as.vector(sigmas[-2]))
  tunegrid = expand.grid(sigma=sigma,C = c(0.75, 1, 2, 4,8,16,32))
  
  mod.svmtune = train(x = datahist[,modensembles],y=as.vector(datahist[,ref]),method = "svmRadial", preProcess = c("center","scale"),importance=T,tuneGrid=tunegrid,trControl = ncvCtrl)
  #estimate final model based on calibrated parameter
  models$SVR = ksvm(x = as.matrix(datahist[,modensembles]),y=as.vector(datahist[,ref]),scaled=T,cross=5,C=mod.svmtune$bestTune$C)

  #Evaluate relative variable importance
  imp.svm <- caret::varImp(mod.svmtune)$importance

  #Cluster stop
  if(parallel == T){stopCluster(cl1)}

  #4. Evaluate and estimate (if est.caltest = T) for cal-test-----

  for(name in namesensembles){
    
    mod_cal = models_cal[[name]]
    mod = models[[name]]
    
    if(class(mod_cal) == "foreccomb_res"){
      pred_cal = mod_cal$Fitted
      pred_test = mod_cal$Forecasts_Test
      pred = mod$Fitted
    }else{
      pred_cal = as.vector(predict(mod_cal,newdata=matrix_cal[,c(modensembles)]))
      pred_test = as.vector(predict(mod_cal,newdata=matrix_test[,c(modensembles)]))
      pred = as.vector(predict(mod,newdata=datahist[,c(modensembles)]))
    }

    #Evaluate cal-val data models
    validation_cal[validation_cal$Model==name,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[datahist[,date] >= CalIni & datahist[,date] <= CalEnd,ref],sim=pred_cal,digits = 5)[c(4,6,9,17,19)]
    validation_test[validation_test$Model==name,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[datahist[,date] >= TestIni & datahist[,date] <= TestEnd,ref],sim=pred_test,digits = 5)[c(4,6,9,17,19)]
    
    #and complete data models
    validation[validation$Model==name,c("RMSE","PBIAS","NSE","R2","KGE")] = gof(obs=datahist[,ref],sim=pred,digits = 5)[c(4,6,9,17,19)]

    #Generate series for HISTORICAL scenario
    datahist[,name] = pred
  }
  
  #Generate series for cal-val data models
  if(est.caltest == T){
    for(name in namesensembles){
      mod_cal = models_cal[[name]]
      if(class(mod_cal) == "foreccomb_res"){
        pred_cal = mod_cal$Fitted
        pred_test = mod_cal$Forecasts_Test
      }else{
        pred_cal = as.vector(predict(mod_cal,newdata=matrix_cal[,c(modensembles)]))
        pred_test = as.vector(predict(mod_cal,newdata=matrix_test[,c(modensembles)]))
      }
      datahist[datahist[,date] >= CalIni & datahist[,date] <= CalEnd,paste0(name,"caltest")] = pred_cal
      datahist[datahist[,date] >= TestIni & datahist[,date] <= TestEnd,paste0(name,"caltest")] = pred_test
    }
  }

  #5. Forecast series (RCP)--------------------------------------------  
  
  #Objects for comb class
  comb.rcp1=foreccomb(datahist[,ref], prediction_matrix=as.matrix(datahist[,modensembles]), newobs = NULL,newpreds = as.matrix(datarcp1[,modensembles]), byrow = FALSE, na.impute = T, criterion = "RMSE")
  
  if(is.null(datarcp2) == F){
    comb.rcp2=foreccomb(datahist[,ref], prediction_matrix=as.matrix(datahist[,modensembles]), newobs = NULL,newpreds = as.matrix(datarcp2[,modensembles]), byrow = FALSE, na.impute = T, criterion = "RMSE")
  }
  
  #Predict for RCP:
  datarcp1[,"SA"] = comb_SA(comb.rcp1)$Forecasts_Test
  datarcp1[,"MED"] = comb_MED(comb.rcp1)$Forecasts_Test  
  datarcp1[,"TA"] = comb_TA(comb.rcp1)$Forecasts_Test
  datarcp1[,"OLS"] = comb_OLS(comb.rcp1)$Forecasts_Test
  datarcp1[,"LAD"] = comb_LAD(comb.rcp1)$Forecasts_Test
  datarcp1[,"CLS"] = comb_CLS(comb.rcp1)$Forecasts_Test
  datarcp1[,"EIG1"] = comb_EIG1(comb.rcp1)$Forecasts_Test
  datarcp1[,"EIG2"] = comb_EIG2(comb.rcp1)$Forecasts_Test

  for(mod in c("BMA","RF","SVR")){
    datarcp1[,mod] = as.vector(predict(models[[mod]],newdata=datarcp1[,c(modensembles)]))
  }  
  
  if(is.null(datarcp2) == F){
    datarcp2[,"SA"] = comb_SA(comb.rcp2)$Forecasts_Test
    datarcp2[,"MED"] = comb_MED(comb.rcp2)$Forecasts_Test
    datarcp2[,"TA"] = comb_TA(comb.rcp2)$Forecasts_Test
    datarcp2[,"OLS"] = comb_OLS(comb.rcp2)$Forecasts_Test
    datarcp2[,"LAD"] = comb_LAD(comb.rcp2)$Forecasts_Test
    datarcp2[,"CLS"] = comb_CLS(comb.rcp2)$Forecasts_Test
    datarcp2[,"EIG1"] = comb_EIG1(comb.rcp2)$Forecasts_Test
    datarcp2[,"EIG2"] = comb_EIG2(comb.rcp2)$Forecasts_Test
    
    for(mod in c("BMA","RF","SVR")){
      datarcp2[,mod] = as.vector(predict(models[[mod]],newdata=datarcp2[,c(modensembles)])) 
    }
  }

  #6. Export results-------------------------------------------

  varimportance=cbind(imp.rf,imp.svm)
  names(varimportance)=c("RF","SVR")
  
  if(is.null(datarcp2) == F){
    results = list(datahist=datahist,datarcp1=datarcp1,datarcp2=datarcp2,validation=validation,validation_cal=validation_cal,validation_test=validation_test,importance=varimportance,models=models)
  }else{
    results = list(datahist=datahist,datarcp1=datarcp1,validation=validation,validation_cal=validation_cal,validation_test=validation_test,importance=varimportance,models=models)
  }
  return(results)
}


















